/*
     Copyright (C) 2006  Georges Da Costa

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along
 with this program; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

//#define TIMESTAMP

#include "toolbox.h"

#include <X11/Xlib.h>
#include <X11/xpm.h>

#include <signal.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h> /* for sprintf */
#include <stdlib.h> /* for malloc */
#include <string.h> /* for strlen */
#include <pwd.h>

#include "gosd.h"
#include "captor.h"
#include "calendar.h"

#include "next.xpm"

#define PIX_HIGH 8
#define PIX_WIDTH 141

Display *display;
int screen;

Pixmap pixmap_table=0;
Pixmap mask_table;

typedef struct font_t {
  unsigned int font;
  unsigned int color;
} font_t;

typedef struct widgets {
  unsigned int x;
  unsigned int y;

  void *data;
  win_t *win;
  void (*handler)(struct widgets *);
  struct widgets* next;
} widgets_t;

widgets_t *widgets=NULL;

extern int max(int a, int b);

void draw_text(int dx,int dy, widgets_t *widget, font_t *font, char *text) {
  //DEBUGIN2("draw_text %s",text);
 gosd_print(widget->win, widget->x+dx,widget->y+dy, font->font, font->color, text);
 //DEBUGOUT2("draw_text %s",text);
}

void sprite(win_t *win, int posx, int posy, int tablex, int tabley, int w, int h) {
  GC gc=DefaultGC(display, screen);

  XSetClipMask(display, gc, mask_table);
  XSetClipOrigin(display, gc, posx-tablex, posy-tabley);
  XCopyArea(display, pixmap_table, gosd_get_buffer(win),
	    gc,
	    tablex, tabley,
	    w, h,
	    posx, posy);

  XSetClipMask(display, gc, None);
}

void jauge(win_t *win, int x, int y, int n) {
  int i;
  i = n*1.41;

  //DEBUGIN("jauge");

 sprite(win, x,y, 0,   0, i, PIX_HIGH);
 sprite(win, x+i,y, i+1, 8, 141-i-1, PIX_HIGH);

 //DEBUGOUT("jauge");
}

void jauge2(win_t *win, int x, int y, int n1, int n2) {
  int i;

  jauge(win, x,y,n2);

  i = n1*1.41;
DEBUGIN("jauge2");

 sprite(win, x,y  , 0,   0, i, PIX_HIGH/2);
 sprite(win, x+i,y, i+1, 8, 141-i-1, PIX_HIGH/2);

DEBUGOUT("jauge2");
}

void normalize(char* st, double size) {
  if(size < 1000) {
    sprintf(st,"%.0fB/s",size);
    return;
  }
  size = size / 1024;
  if(size < 1000) {
    sprintf(st,"%.1fkB/s",size);
    return;
  }
  size = size / 1024;

  sprintf(st,"%.2fMB/s",size);
  return;
}

void event_sound(unsigned int a, unsigned int b, unsigned int c, void *widget) {
  int x=a;
  
  if(c == 4) x = -1;
  else if(c == 5) x = -3;
  else x = x/1.41;

  if(x >= 0) {
    set_vol(x); set_pcm(x);
  } else {
    set_vol(get_vol()+(x+2)*5);
    set_pcm(get_pcm()+(x+2)*5);
  }
}

void event_cpufreq(unsigned int a, unsigned int b, unsigned int c, void *widget) {
  set_cpufreq((get_cpufreq()+1)%3);
}

void draw_widgets(win_t *w, void *data) {
  widgets_t* current = widgets;
DEBUGIN2("maj %p",w)

  if(gosd_lock_display() == -1) {
DEBUGOUT("maj locked")
    return;
  }

  gosd_put_background(w);

  while(current != NULL) {
    if(current->win == w) {
      unsigned int x,y;
      x=current->x;
      y=current->y;
      current->handler(current);
    }
    current = current->next;
  }

  gosd_update_display(w);
  gosd_release_display();
DEBUGOUT2("maj %p",w)
}

void handler_vol(widgets_t *widget) {
  int x=widget->x;
  int y=widget->y;

  sprite(widget->win, x,y, 60,17, 12,14);
  jauge2(widget->win, x+20, y+3, get_vol(), get_pcm());
}

void handler_load(widgets_t *widget) {
  char buffer[20];
  int x=widget->x;
  int y=widget->y;
  jauge(widget->win, x, y+1 , get_load());
  if(get_temperature() > 0) {
    sprintf(buffer,"%dC",get_temperature());
    draw_text(3+PIX_WIDTH,0,widget, (font_t*) widget->data, buffer);
  }
}

void handler_cpufreq(widgets_t *widget) {
  int x=widget->x;
  int y=widget->y;
  sprite(widget->win, x,y, (2-get_cpufreq())*20, 17, 20, 16);
}

void handler_time(widgets_t *widget) {
  char status[8];
  strftime(status, 8, "%H : %M", get_time());
  draw_text(0,0,widget,(font_t*) widget->data, status);
}

void handler_date(widgets_t *widget) {
  char status[13];
  strftime(status, 13, "%d %b %Y", get_time());
  draw_text(0,0,widget,(font_t*) widget->data, status);

}

void handler_network(widgets_t *widget) {
  char status_download[10];
  char status_upload[10];
  if(get_upload() != 0 || get_download() != 0) {
    normalize(status_download, get_download());
    normalize(status_upload, get_upload());

    draw_text(0,0,widget,(font_t*) widget->data, status_download);
    draw_text(77,0,widget,(font_t*) widget->data, status_upload);
  }
}

void handler_ac(widgets_t *widget) {
  int x=widget->x;
  int y=widget->y;
  if(get_ac())
    sprite(widget->win, x+10,y, 89,18,16,12);
  else 
    sprite(widget->win, x,y, 114,18,26,12);
}

void handler_battery(widgets_t *widget) {
  int x=widget->x;
  int y=widget->y;
  if(get_rate_battery() != 0) {
    if(!get_ac()) {
      char buffer[20];
      int h = get_battery()/get_rate_battery();
      int m = (60*get_battery())/get_rate_battery() - 60*h;
      /*
      if(m>=10)
	sprintf(buffer,"%2dh%d",h,m);
      else
	sprintf(buffer,"%dh0%d",h,m);
      */
      sprintf(buffer," %dh%.2d",h,m);
      draw_text(141,0,widget,(font_t*) widget->data, buffer);
    }
    jauge(widget->win, x, y ,get_battery());
  }
}

void handler_calendar(widgets_t *widget) {
  calendar(widget->win, widget->data);
}

void handler_null(widgets_t *widget) {
}

#define SIZE(w,h) if(sizex<x+w) sizex=x+w; if(sizey<y+h) sizey=y+h

void finalize_window(int pos, int posx, int posy, int delta,
		     unsigned int sizex, unsigned int sizey, win_t *w) {
  unsigned int px= ScreenOfDisplay(display,screen)->width;
  unsigned int py= ScreenOfDisplay(display,screen)->height;
  if(pos == 2) {
    posx = px/2-delta;
    posy = py - sizey-1;
  } else if(pos == 3) {
    posx= px/2-delta;
    posy= 1;
  } else if(pos == 4) {
    posx=1;
    posy=py/2-delta;
  } else if(pos == 5) {
    posx=px-sizex-1;
    posy=py/2-delta;
  }

  gosd_position(posx,posy,w);
  gosd_size(sizex, sizey,w);
}

void init_conf(int argc, char **argv) {
  FILE* conf;
  char buffer[100];
  char argument[100];
  unsigned int sizex=0;
  unsigned int sizey=0;
  char *conf_file;
  struct passwd *pwd;
  int pos=0;
  int delta=0;
  unsigned int posx=0, posy=0;
  win_t* w=NULL;

  if(argc == 1) {
    pwd = getpwuid(getuid());
    conf_file=malloc(28+strlen(pwd->pw_dir));
    sprintf(conf_file,"%s/.config/light-monitor.conf", pwd->pw_dir);
    conf = fopen(conf_file,"r");
    free(conf_file);
    if(conf == NULL) {
      conf = fopen("/etc/light-monitor.conf","r");
      if(conf==NULL) {
	printf("No config file :\n~/.config/light-monitor.conf\n");
	exit(1);
      }
    }
  } else {
    conf = fopen(argv[1],"r");
    if(conf == NULL) {
      printf("File (%s) not found\n", argv[1]);
      exit(1);
    }
  }
  while(!feof(conf)) {
    buffer[0]='\0';
    fgets(buffer, 99, conf);

    if(!strncmp(buffer, "[", 1)) {
      int l;
      char *title;

      if(w != NULL) {
	finalize_window(pos, posx, posy, delta, sizex, sizey, w);
	pos=0;
	posx=0;
	posy=0;
	delta=0;
	sizex=0;
	sizey=0;
      }


      l=strlen(buffer)-3;
      title=malloc(l+1);
      strncpy(title, buffer+1, l);
      title[l]='\0';
      w=gosd_new_window(0, 0, 1, 1, title);

      if(pixmap_table == 0)
	XpmCreatePixmapFromData(display, gosd_get_window(w), next_xpm,
				&pixmap_table, &mask_table, NULL);
      gosd_set_handler(&draw_widgets, NULL, w);

    } else if(!strncmp(buffer, "pos",3)) {
      sscanf(buffer+4, "%u %u\n", &posx, &posy);
    } else if (!strncmp(buffer, "bottom", 6)) {
      sscanf(buffer+6, "%u\n", &delta);
      pos=2;
    } else if (!strncmp(buffer, "top", 3)) {
      sscanf(buffer+3, "%u\n", &delta);
      pos=3;
    } else if (!strncmp(buffer, "left", 4)) {
      sscanf(buffer+4, "%u\n", &delta);
      pos=4;
    } else if (!strncmp(buffer, "right", 5)) {
      sscanf(buffer+5, "%u\n", &delta);
      pos=5;
    } else if (!strncmp(buffer, "color",5)) {
      short unsigned int r,g,b;
      int n;
      sscanf(buffer+6, "%d %hu %hu %hu\n", &n,
	     &r, &g, &b);
      gosd_add_color(w, n, r, g, b);

    } else if (!strncmp(buffer, "font",4)) {
      int n; double size;
      sscanf(buffer+5, "%d %lf %s\n", &n, &size, buffer);

      gosd_add_font(w, n, size, buffer);

    } else if (!strncmp(buffer, "calendar",8)) {
      unsigned int x,y,f1,f2,f3,c1,c2, deca;
      widgets_t* old = widgets;
      widgets = malloc(sizeof(struct widgets));
      widgets->next = old;

 int cal_w=196;
 int cal_h=160;

      sscanf(buffer+8, "%u %u %u %u %u %u %u %u\n",
	                &x,&y,&f1,&f2,&f3,&c1,&c2,&deca);

      widgets->x = x;
      widgets->y = y;
      widgets->win = w;
      widgets->data = new_calendar(w,f1,f2,f3,c1,c2,deca);
      widgets->handler = handler_calendar;
      gosd_add_button(151, 14, 14, 20, 1, prev_month,   widgets->data,w);
      gosd_add_button(167, 14, 14, 20, 1, current_month,widgets->data,w);
      gosd_add_button(181, 14, 14, 20, 1, next_month,   widgets->data,w);

      SIZE(cal_w,cal_h);

    } else if (!strncmp(buffer, "widget",6)) {
      unsigned int x,y,f,color;
      widgets_t* old = widgets;
      font_t* font = malloc(sizeof(font_t));

      argument[0]='\0';
      sscanf(buffer+7, "%u %u %u %u %s %s\n", &x, &y, &f, &color,
	                                      buffer, argument);
      widgets = malloc(sizeof(struct widgets));
      widgets->x = x;
      widgets->y = y;
      font->font = f;
      font->color = color;
      widgets->data = (void*) font;
      widgets->win = w;

      if(!strncmp(buffer, "vol",3)) {
	widgets->handler = handler_vol;
	gosd_add_button(x+20, y, 150, 10, 1, event_sound,(void*)widgets, w);
	gosd_add_button(x+20, y, 150, 10, 4, event_sound,(void*)widgets, w);
	gosd_add_button(x+20, y, 150, 10, 5, event_sound,(void*)widgets, w);
	SIZE(159,14);
      }
      else if(!strncmp(buffer, "load",4)) {
	widgets->handler = handler_load;
	SIZE(141+3+gosd_text_width(w,f,"88C"),max(9,gosd_text_height(w,f,"88C")));
	if(argument[0] != '\0')
	  init_temperature(argument);
      }
      else if(!strncmp(buffer, "cpufreq",7)) {
	widgets->handler = handler_cpufreq;
	gosd_add_button(x, y, 20, 15, 1, event_cpufreq,(void*)widgets, w);
	SIZE(20,15);
	if(argument[0] != '\0')
	  init_cpufreq(argument);
      }
      else if(!strncmp(buffer, "time",4)) {
	widgets->handler = handler_time;
	SIZE(gosd_text_width(w,f,"22 : 43"),gosd_text_height(w,f,"22 : 33"));
      }
      else if(!strncmp(buffer, "date",4)) {
	widgets->handler = handler_date;
	SIZE(gosd_text_width(w,f,"22 Jan 2008"),gosd_text_height(w,f,"22 Jan 2008"));
      }
      else if(!strncmp(buffer, "network",7)) {
	widgets->handler = handler_network;
	SIZE(77+gosd_text_width(w,f,"888.2MB/s"),gosd_text_height(w,f,"88 B/s"));
      }
      else if(!strncmp(buffer, "ac",2)) {
	widgets->handler = handler_ac;
	SIZE(26,12);
	if(argument[0] != '\0')
	  init_ac(argument);
      }
      else if(!strncmp(buffer, "battery",7)) {
	widgets->handler = handler_battery;
	SIZE(141+gosd_text_width(w,f," 2h30"),max(9,gosd_text_height(w,f," 2h30")));
	if(argument[0] != '\0')
	  init_battery(argument);
      }
      else
	widgets->handler = handler_null;
      widgets->next = old;
    }
  }
  fclose(conf);

  finalize_window(pos, posx, posy, delta, sizex, sizey, w);
  
}

int main(int argc, char **argv) {

  if(argc==2 && !strcmp(argv[1], "--version")) {
    printf("Light-monitor version 1.7\nGPLv2 license\n");
    exit(0);
  }

SHOWTIME("main")

  gosd_init();

SHOWTIME("gosd init")

  display = gosd_get_display();
  screen = DefaultScreen (display);

SHOWTIME("get X info")

  init_conf(argc, argv);

SHOWTIME("init conf")

  init_captor();

SHOWTIME("set handler")

  gosd_add_timeout(&maj_captors, NULL);

SHOWTIME("add timeout")

  gosd_loop();

  return 0;
}
